import json
from io import BytesIO
from typing import BinaryIO, Iterable, Tuple, Union

import numpy as np
from PIL import Image

from brainframe.api.bf_codecs import image_utils
from .base_stub import BaseStub, DEFAULT_TIMEOUT


class StorageStubMixin(BaseStub):
    """Provides stubs to call APIs for managing binary blob storage."""

    def get_storage_data(self, storage_id,
                         timeout=DEFAULT_TIMEOUT) -> Tuple[bytes, str]:
        """Returns the data with the given storage ID.

        :param storage_id: The ID of the storage object to get
        :param timeout: The timeout to use for this request
        :return: The data and MIME type of that data
        """
        req = f"/api/storage/{storage_id}"
        resp = self._get(req, timeout)

        return resp.content, resp.headers["Content-Type"]

    def get_storage_data_as_image(self, storage_id,
                                  timeout=DEFAULT_TIMEOUT) -> np.ndarray:
        """Gets the data with the given storage ID and attempts to load it as
        an image with OpenCV.

        :param storage_id: The ID of the storage object to get
        :param timeout: The timeout to use for this request
        :return: A numpy array in OpenCV format
        """
        data, _ = self.get_storage_data(storage_id, timeout=timeout)
        return image_utils.decode(data)

    def new_storage(self, data: Union[bytes, BinaryIO, Iterable],
                    mime_type: str,
                    timeout=DEFAULT_TIMEOUT) -> int:
        """Stores the given data.

        :param data: The data to store, either as bytes or as a file-like
        :param mime_type: The MIME type of the data
        :param timeout: The timeout to use for this request
        :return: The storage ID
        """
        req = r"/api/storage"

        storage_id_json = self._post(req, timeout, data, mime_type).content
        return json.loads(storage_id_json)

    def new_storage_as_image(self, data: bytes,
                             timeout=DEFAULT_TIMEOUT) -> int:
        """Stores the given image data, and inspects it to figure out what the
        MIME type of the data.

        :param data: The image data to store
        :param timeout: The timeout to use for this request
        :return: The storage ID
        """
        try:
            pil_image = Image.open(BytesIO(data))

            if pil_image.format == "JPEG":
                mime_type = "image/jpeg"
            elif pil_image.format == "PNG":
                mime_type = "image/png"
            else:
                mime_type = "application/octet-stream"
        except IOError:
            # TODO(Tyler Compton): This failure mode is stupid but I don't want
            # two ways for this call to fail when an invalid image is passed to
            # it. Figure out a better way.
            mime_type = "application/octet-stream"

        return self.new_storage(data, mime_type, timeout=timeout)

    def delete_storage(self, storage_id, timeout=DEFAULT_TIMEOUT):
        """Deletes the storage object with the given ID. Deleting storage
        objects that are in use, like for an encoding, is not recommended.

        :param storage_id: The ID of the storage object to delete
        :param timeout: The timeout to use for this request
        """
        req = f"/api/storage/{storage_id}"

        self._delete(req, timeout)



from typing import Dict, List

import numpy as np
import json

from brainframe.api.bf_codecs import Detection, image_utils
from .base_stub import BaseStub, DEFAULT_TIMEOUT


class ProcessImageStubMixIn(BaseStub):
    """Provides stubs to call APIs that run processing on a single frame."""

    def process_image(self, img_bgr: np.ndarray,
                      capsule_names: List[str],
                      option_vals: Dict[str, Dict[str, object]],
                      timeout=DEFAULT_TIMEOUT) \
            -> List[Detection]:
        """Process a single image using the given configuration.
        :param img_bgr: The image to process
        :param capsule_names: The capsule names to enable while processing the
            image
        :param option_vals: Capsule option values, where the key is a capsule
            name and the value is a dict with key-value pairs for each option
            and its corresponding value. Any specified options will override
            the global option values for that capsule
        :param timeout: The timeout to use for this request
        :return: All detections in the image
        """
        req = f"/api/process_image"

        metadata = {
            "plugins": capsule_names,
            "options": option_vals
        }
        metadata = json.dumps(metadata)

        # Encode the image
        img_bytes = image_utils.encode("jpeg", img_bgr)

        files = {
            "image": ("image.jpg",
                      img_bytes,
                      "image/jpeg"),
            "metadata": ("metadata.json",
                         metadata.encode("utf-8"),
                         "application/json")}

        resp = self._post_multipart(req, timeout, files)
        return [Detection.from_dict(d) for d in resp]



from typing import List

from brainframe.api.bf_codecs import User
from .base_stub import BaseStub, DEFAULT_TIMEOUT


class UserStubMixin(BaseStub):
    """Provides stubs for calling APIs related to user management."""

    def get_user(self, user_id,
                 timeout=DEFAULT_TIMEOUT) -> User:
        """Gets the user with the given ID.

        :param user_id: The ID of the user to get
        :param timeout: The timeout to use for this request
        :return: The user
        """
        req = f"/api/users/{user_id}"
        data, _ = self._get_json(req, timeout)

        return User.from_dict(data)

    def get_users(self, timeout=DEFAULT_TIMEOUT) -> List[User]:
        """Gets all users.

        :param timeout: The timeout to use for this request
        :return: Users
        """
        req = f"/api/users"
        data, _ = self._get_json(req, timeout)

        return [User.from_dict(u) for u in data]

    def set_user(self, user, timeout=DEFAULT_TIMEOUT) -> User:
        """Creates or updates a user. Only admin users may make this request.

        A new user is created if the given user's ID is None.

        :param user: The user to create or update
        :param timeout: The timeout to use for this request
        :return: Created or updated user
        """
        req = f"/api/users"
        data = self._post_codec(req, timeout, user)

        return User.from_dict(data)

    def delete_user(self, user_id, timeout=DEFAULT_TIMEOUT):
        """Deletes a user. Only admin users may make this request.

        :param user_id: The ID of the user to delete
        :param timeout: The timeout to use for this request
        """
        req = f"/api/users/{user_id}"
        self._delete(req, timeout)



from dataclasses import dataclass

from .base_codecs import Codec


@dataclass
class CloudTokens(Codec):
    """OAuth2 tokens for a BrainFrame Cloud user"""

    access_token: str
    """Provides access to the API"""

    refresh_token: str
    """Used to refresh the access token when it expires"""

    def to_dict(self) -> dict:
        return dict(self.__dict__)

    @staticmethod
    def from_dict(d: dict) -> "CloudTokens":
        return CloudTokens(
            access_token=d["access_token"],
            refresh_token=d["refresh_token"],
        )


@dataclass
class CloudUserInfo(Codec):
    """Information on a BrainFrame Cloud user"""

    sub: str
    """A unique identifier for the user"""

    email: str
    """The user's email address"""

    def to_dict(self):
        return dict(self.__dict__)

    @staticmethod
    def from_dict(d: dict) -> "CloudUserInfo":
        return CloudUserInfo(
            sub=d["sub"],
            email=d["email"],
        )



from typing import List, Optional

from brainframe.api.bf_codecs import ZoneAlarm
from .base_stub import BaseStub, DEFAULT_TIMEOUT


class ZoneAlarmStubMixin(BaseStub):
    """Provides stubs for calling APIs related to getting zone alarms."""

    def get_zone_alarm(self, alarm_id,
                       timeout=DEFAULT_TIMEOUT) -> ZoneAlarm:
        """Gets the zone alarm with the given ID.

        :param alarm_id: The ID of the alarm to get
        :param timeout: The timeout to use for this request
        :return: The alarm
        """
        req = f"/api/zone_alarms/{alarm_id}"
        data, _ = self._get_json(req, timeout)

        return ZoneAlarm.from_dict(data)

    def get_zone_alarms(self, stream_id: Optional[int] = None,
                        zone_id: Optional[int] = None,
                        timeout=DEFAULT_TIMEOUT) -> List[ZoneAlarm]:
        """Gets all zone alarms that fit the given filters.

        :param stream_id: If supplied, only zone alarms that are for a zone in
            this stream will be returned
        :param zone_id: if supplied, only zone alarms attached to this zone
            will be returned
        :param timeout: The timeout to use for this request
        :return: Zone alarms
        """
        req = f"/api/zone_alarms"

        params = {}
        if stream_id is not None:
            params["stream_id"] = stream_id
        if zone_id is not None:
            params["zone_id"] = zone_id

        data, _ = self._get_json(req, timeout, params=params)

        return [ZoneAlarm.from_dict(a) for a in data]

    def set_zone_alarm(self, alarm: ZoneAlarm,
                       timeout=DEFAULT_TIMEOUT) -> ZoneAlarm:
        """Creates or updates a zone alarm.

        A new zone alarm is created if the given zone alarm's ID is None.

        :param alarm: The zone alarm to create or update
        :param timeout: The timeout to use for this request
        :return: Created or updated zone alarm
        """
        req = f"/api/zone_alarms"
        data = self._post_codec(req, timeout, alarm)

        return ZoneAlarm.from_dict(data)

    def delete_zone_alarm(self, alarm_id,
                          timeout=DEFAULT_TIMEOUT):
        """Deletes the zone alarm with the given ID.

        :param alarm_id: The ID of the zone alarm to delete
        :param timeout: The timeout to use for this request
        """
        req = f"/api/zone_alarms/{alarm_id}"
        self._delete(req, timeout)



from .base_stub import BaseStub, DEFAULT_TIMEOUT


class AnalysisStubMixin(BaseStub):
    """Provides stubs for calling APIs that control analysis on streams."""

    def start_analyzing(self, stream_id,
                        timeout=DEFAULT_TIMEOUT):
        """Starts analysis on this stream.

        :param stream_id: The ID of the stream to start analysis on
        :param timeout: The timeout to use for this request
        """
        req = f"/api/streams/{stream_id}/analyze"
        self._put_json(req, timeout, 'true')

    def stop_analyzing(self, stream_id,
                       timeout=DEFAULT_TIMEOUT):
        """Stops analysis on this stream.

        :param stream_id: The ID of the stream to stop analysis on
        :param timeout: The timeout to use for this request
        :return: True or False if the server was able to start analysis on that
            stream. It could fail because: unable to start stream, or license
            restrictions.
        """
        req = f"/api/streams/{stream_id}/analyze"
        resp = self._put_json(req, timeout, 'false')
        return resp

    def check_analyzing(self, stream_id,
                        timeout=DEFAULT_TIMEOUT) -> bool:
        """Check if this stream is being analyzed

        :param stream_id: The ID of the stream to check
        :param timeout: The timeout to use for this request
        :return: True or False if the stream is being analyzed
        """
        req = f"/api/streams/{stream_id}/analyze"
        resp, _ = self._get_json(req, timeout)
        return resp



from typing import Dict, List, Optional

import json

from brainframe.api.bf_codecs import StreamConfiguration
from .base_stub import BaseStub, DEFAULT_TIMEOUT


class StreamStubMixin(BaseStub):
    """Provides stubs for calling stream-related APIs and managing
    stream-related client objects.
    """

    def __init__(self):
        # These are evaluated lazily
        self._status_receiver = None

    def get_status_receiver(self) -> "StatusReceiver":
        """Returns a singleton StatusReceiver object"""
        from brainframe.api import StatusReceiver
        if self._status_receiver is None \
                or not self._status_receiver.is_running:
            self._status_receiver = StatusReceiver(self)
        return self._status_receiver

    def get_stream_configuration(self, stream_id, timeout=DEFAULT_TIMEOUT) \
            -> StreamConfiguration:
        """Gets the StreamConfiguration with the given ID.

        :param stream_id: The ID of the stream configuration to get
        :param timeout: The timeout to use for this request
        :return: The stream configuration
        """
        req = f"/api/streams/{stream_id}"
        data, _ = self._get_json(req, timeout)

        return StreamConfiguration.from_dict(data)

    def get_stream_configurations(self, premises_id=None,
                                  timeout=DEFAULT_TIMEOUT) \
            -> List[StreamConfiguration]:
        """Get all StreamConfigurations that currently exist.

        :return: [StreamConfiguration, StreamConfiguration, ...]
        """
        req = "/api/streams"
        params = {"premises_id": premises_id} if premises_id else None
        data, _ = self._get_json(req, timeout, params=params)

        configs = [StreamConfiguration.from_dict(d) for d in data]
        return configs

    def set_stream_configuration(self, stream_configuration,
                                 timeout=DEFAULT_TIMEOUT) \
            -> Optional[StreamConfiguration]:
        """Update an existing stream configuration or create a new one. If
        creating a new one, the stream_configuration.id will be None.

        :param stream_configuration: StreamConfiguration
        :param timeout: The timeout to use for this request
        :return: StreamConfiguration, initialized with an ID
        """
        req = "/api/streams"
        data = self._post_codec(req, timeout, stream_configuration)
        config = StreamConfiguration.from_dict(data)
        return config

    # TODO: Remove this long timeout when this endpoint is better optimized
    def delete_stream_configuration(self, stream_id,
                                    timeout=120):
        """Deletes a stream configuration with the given ID. Also stops
        analysis if analysis was running and closes the stream.

        :param stream_id: The ID of the stream to delete
        :param timeout: The timeout to use for this request
        """
        req = f"/api/streams/{stream_id}"
        self._delete(req, timeout)

    def get_stream_url(self, stream_id,
                       timeout=DEFAULT_TIMEOUT) -> str:
        """Gets the URL that the stream is available at.

        :param stream_id: The ID of the stream to get a URL for
        :param timeout: The timeout to use for this request
        :return: The URL
        """
        req = f"/api/streams/{stream_id}/url"
        url, _ = self._get_json(req, timeout)
        return url

    def get_runtime_options(self, stream_id: int,
                            timeout=DEFAULT_TIMEOUT) -> Dict[str, object]:
        """Returns the runtime options for the stream with the given ID. This
        can also be read from a StreamConfiguration object.

        :param stream_id: The ID of the stream to get runtime options from
        :param timeout: The timeout to use for this request
        :return: Runtime options
        """
        req = f"/api/streams/{stream_id}/runtime_options"
        runtime_options, _ = self._get_json(req, timeout)

        return runtime_options

    def set_runtime_option_vals(self, stream_id: int,
                                runtime_options: Dict[str, object],
                                timeout=DEFAULT_TIMEOUT):
        """Sets a stream's runtime options to the given values.

        :param stream_id: The ID of the stream to set runtime options for
        :param runtime_options: The runtime options
        :param timeout: The timeout to use for this requestz
        """
        req = f"/api/streams/{stream_id}/runtime_options"
        runtime_options_json = json.dumps(runtime_options)
        self._put_json(req, timeout, runtime_options_json)

    def close(self):
        if self._status_receiver is not None:
            self._status_receiver.close()
            self._status_receiver = None



from typing import Tuple
from time import sleep, time

from requests.exceptions import ConnectionError, ReadTimeout

from .bf_errors import UnauthorizedError, UnknownError, ServerNotReadyError
from . import stubs
from .stubs.base_stub import DEFAULT_TIMEOUT


class BrainFrameAPI(stubs.AlertStubMixin,
                    stubs.AnalysisStubMixin,
                    stubs.IdentityStubMixin,
                    stubs.CapsuleStubMixin,
                    stubs.StreamStubMixin,
                    stubs.ZoneStatusStubMixin,
                    stubs.ZoneStubMixin,
                    stubs.StorageStubMixin,
                    stubs.ZoneAlarmStubMixin,
                    stubs.ProcessImageStubMixIn,
                    stubs.EncodingStubMixIn,
                    stubs.PremisesStubMixin,
                    stubs.UserStubMixin,
                    stubs.LicenseStubMixIn,
                    stubs.CloudTokensStubMixin,
                    stubs.CloudUsersStubMixIn,
                    stubs.OAuth2StubMixIn):
    """Provides access to BrainFrame API endpoints."""

    def __init__(self, server_url=None, credentials: Tuple[str, str] = None):
        """
        :param server_url: The URL of the BrainFrame instance to connect to. If
            None, it needs to be set later with set_url before use
        :param credentials: The username and password as a tuple. Used to
            authenticate with the server. If None, no authentication
            information will be provided.
        """
        super().__init__()
        self._server_url = server_url
        self._credentials = credentials

    def version(self, timeout=DEFAULT_TIMEOUT) -> str:
        """
        :return: The current BrainFrame version in the format X.Y.Z
        """
        req = f"/api/version"

        resp, _ = self._get_json(req, timeout)
        return resp

    def wait_for_server_initialization(self, timeout: float = None):
        """Waits for the server to be ready to handle requests.

        :param timeout: The maximum amount of time, in seconds, to wait for the
            server to start. If None, this method will wait indefinitely.
        """
        start_time = time()
        while True:
            if timeout is not None and time() - start_time > timeout:
                raise TimeoutError("The server did not start in time!")

            try:
                # Test connection to server
                self.version()
            except (ConnectionError, ConnectionRefusedError, ReadTimeout,
                    UnauthorizedError, ServerNotReadyError):
                # Server not started yet or there is a communication
                # error
                pass
            except UnknownError as exc:
                if exc.status_code not in [502]:
                    raise
            else:
                break

            # Prevent busy loop
            sleep(.1)

    def close(self):
        """Clean up the API. It may no longer be used after this call."""
        stubs.StreamStubMixin.close(self)



from brainframe.api.bf_codecs import OAuth2Info
from .base_stub import BaseStub, DEFAULT_TIMEOUT


class OAuth2StubMixIn(BaseStub):
    """Provides stubs for calling APIs related to OAuth2"""

    def get_oauth2_info(self, timeout: float = DEFAULT_TIMEOUT) -> OAuth2Info:
        """Gets the necessary information to start an OAuth2 on behalf of the server.

        :param timeout: The timeout to use for this request
        :return: The OAuth2 info
        """
        req = "/api/oauth2_info"
        data, _ = self._get_json(req, timeout)
        return OAuth2Info.from_dict(data)



from collections import Counter
from typing import ClassVar, List, Optional

from dataclasses import dataclass, field

from .base_codecs import Codec
from .alarm_codecs import Alert, ZoneAlarm
from .detection_codecs import Detection


@dataclass
class Zone(Codec):
    """The definition for a zone. It is a non-convex polygon or a line."""
    FULL_FRAME_ZONE_NAME: ClassVar[str] = "Screen"
    """The name of the default full-frame zone that exists on every stream"""

    name: str
    """A friendly name for the zone"""

    stream_id: int
    """The ID of the stream this zone is in"""

    coords: List[List[int]]
    """Coordinates that define the region the zone occupies. It is a list
    of lists which are two elements in size. The coordinates are in pixels
    where the top left of the frame is [0, 0].
    
    Example: [[0, 0], [10, 10], [100, 500], [0, 500]]
    """

    alarms: List[ZoneAlarm] = field(default_factory=list)
    """All alarms that are attached to the zone"""

    id: Optional[int] = None
    """A unique identifier"""

    def get_alarm(self, alarm_id) -> Optional[ZoneAlarm]:
        """
        :param alarm_id: The ID of the alarm to search in the alarm list for
        :return: The alarm with the given ID, or None if no alarm with that ID
            exists in this zone
        """
        for alarm in self.alarms:
            if alarm.id == alarm_id:
                return alarm
        return None

    def to_dict(self):
        d = dict(self.__dict__)
        d["alarms"] = [alarm.to_dict() for alarm in self.alarms]
        return d

    @staticmethod
    def from_dict(d):
        alarms = [ZoneAlarm.from_dict(alarm_d) for alarm_d in d["alarms"]]
        return Zone(name=d["name"],
                    id=d["id"],
                    alarms=alarms,
                    stream_id=d["stream_id"],
                    coords=d["coords"])


@dataclass
class ZoneStatus(Codec):
    """The current status of everything going on inside a zone.
    """

    zone: Zone
    """The zone that this status pertains to"""

    tstamp: float
    """The time at which this ZoneStatus was created as a Unix timestamp in
    seconds
    """

    total_entered: dict
    """A dict of key-value pairs indicating how many objects have exited the
    zone. The key is the class name, and the value is the count.
    """

    total_exited: dict
    """A set of key-value pairs indicating how many objects have exited the
    zone. The key is the object type, and the value is the count.
    """

    within: List[Detection]
    """A list of all detections within the zone"""

    entering: List[Detection]
    """A list of all detections that entered the zone this frame"""

    exiting: List[Detection]
    """A list of all detections that have exited the zone this frame"""

    alerts: List[Alert]
    """A list of all active alerts for the zone at this frame"""

    @property
    def detection_within_counts(self) -> dict:
        """The current count of each class type detected in the video.

        :returns: A dict whose keys are class names and whose values are the
            count for that class name
        """
        counter = Counter([det.class_name for det in self.within])
        return counter

    def to_dict(self):
        d = dict(self.__dict__)
        d["zone"] = self.zone.to_dict()
        d["within"] = [det.to_dict() for det in self.within]
        d["entering"] = [det.to_dict() for det in self.entering]
        d["exiting"] = [det.to_dict() for det in self.exiting]
        d["alerts"] = [alert.to_dict() for alert in self.alerts]
        return d

    @staticmethod
    def from_dict(d):
        zone = Zone.from_dict(d["zone"])
        within = [Detection.from_dict(det) for det in d["within"]]
        entering = [Detection.from_dict(det) for det in d["entering"]]
        exiting = [Detection.from_dict(det) for det in d["exiting"]]
        alerts = [Alert.from_dict(alert) for alert in d["alerts"]]
        return ZoneStatus(zone=zone,
                          tstamp=d["tstamp"],
                          total_entered=d["total_entered"],
                          total_exited=d["total_exited"],
                          within=within,
                          entering=entering,
                          exiting=exiting,
                          alerts=alerts)



from .base_codecs import Codec

from .alarm_codecs import Alert, ZoneAlarm
from .condition_codecs import (
    ZoneAlarmCountCondition,
    ZoneAlarmRateCondition,
    IntersectionPointType,
)
from .config_codecs import StreamConfiguration
from .identity_codecs import Identity, Encoding
from .detection_codecs import Attribute, Detection
from .zone_codecs import Zone, ZoneStatus
from .capsule_codecs import (
    CapsuleOption,
    Capsule,
    NodeDescription,
)
from .premises_codecs import Premises
from .user_codecs import User
from .license_codecs import (
    LicenseTerms,
    LicenseInfo,
    DATE_FORMAT,
)
from .cloud_codecs import CloudTokens, CloudUserInfo
from .oauth2_codecs import OAuth2Info
from .sorting import SortOptions, Ordering



from datetime import date, datetime
from typing import Optional
from enum import Enum

from dataclasses import dataclass

from .base_codecs import Codec


@dataclass
class LicenseTerms(Codec):
    """The terms of the currently active license."""

    online_checkin: bool
    """If True, the server will check in with a remote licensing server to
    verify license terms.
    """

    max_streams: int
    """The maximum number of streams that may have analysis enabled at any
    given time.
    """

    journal_max_allowed_age: float
    """The maximum amount of time in seconds that the server may hold data
    in the journal for.
    """

    expiration_date: Optional[date]
    """The date that this license expires, or None if the license is
    perpetual.
    """

    def to_dict(self) -> dict:
        d = dict(self.__dict__)
        if self.expiration_date is not None:
            d["expiration_date"] = self.expiration_date.isoformat()
        return d

    @staticmethod
    def from_dict(d: dict) -> "LicenseTerms":
        expiration_date = None
        if d["expiration_date"] is not None:
            # TODO: Switch to date.fromisoformat when we're on Python >= 3.7
            expiration_date = datetime.strptime(
                d["expiration_date"], DATE_FORMAT).date()

        terms = LicenseTerms(
            online_checkin=d["online_checkin"],
            max_streams=d["max_streams"],
            journal_max_allowed_age=d["journal_max_allowed_age"],
            expiration_date=expiration_date,
        )
        return terms


@dataclass
class LicenseInfo(Codec):
    """Information on the licensing status of the server"""

    class State(Enum):
        VALID = "valid"
        """A valid license is loaded, features should be enabled"""
        INVALID = "invalid"
        """A license was provided, but did not pass validation"""
        EXPIRED = "expired"
        """A license was provided, but it has expired"""
        MISSING = "missing"
        """No license was provided"""

    state: State
    """The licensing state of the server."""

    terms: Optional[LicenseTerms]
    """The active license terms of the server, or None if no license is
    loaded.
    """

    def to_dict(self) -> dict:
        return {
            "state": self.state.value,
            "terms": self.terms.to_dict() if self.terms else None,
        }

    @staticmethod
    def from_dict(d: dict) -> "LicenseInfo":
        terms = None
        if d["terms"] is not None:
            terms = LicenseTerms.from_dict(d["terms"])

        return LicenseInfo(
            state=LicenseInfo.State(d["state"]),
            terms=terms,
        )


DATE_FORMAT = "%Y-%m-%d"
"""ISO 8601 date format used by the API"""



"""Some utility functions to make encoding and decoding images with Pillow
easy.
"""
from io import BytesIO

import numpy as np
from PIL import Image


def decode(img_bytes: bytes) -> np.ndarray:
    rgb_img = Image.open(BytesIO(img_bytes))

    # Load as numpy array
    bgr_arr = np.asarray(rgb_img, dtype=np.uint8)

    # Convert from RGB to BGR
    bgr_arr = flip_channels(bgr_arr)

    # Return a fresh copy instead of a "memoryview" object...
    return bgr_arr.copy()


def encode(format: str, image_bgr_arr: np.ndarray):
    img_rgb_arr = flip_channels(image_bgr_arr)
    image = Image.fromarray(img_rgb_arr)
    img_bytes = BytesIO()

    image.save(img_bytes, format=format)
    img_bytes = img_bytes.getvalue()

    return img_bytes


def flip_channels(img_arr):
    return img_arr[..., ::-1]



from typing import List

from brainframe.api.bf_codecs import Zone
from .base_stub import BaseStub, DEFAULT_TIMEOUT


class ZoneStubMixin(BaseStub):
    """Provides stubs for calling APIs to get, set, and delete zones."""

    def get_zones(self, stream_id=None,
                  timeout=DEFAULT_TIMEOUT) -> List[Zone]:
        """Gets all zones.

        :param stream_id: If set, only zones in the stream with this ID are
            returned
        :param timeout: The timeout to use for this request
        :return: Zones
        """
        req = "/api/zones"
        params = {"stream_id": stream_id} if stream_id else None
        data, _ = self._get_json(req, timeout, params=params)
        zones = [Zone.from_dict(j) for j in data]
        return zones

    def get_zone(self, zone_id,
                 timeout=DEFAULT_TIMEOUT) -> Zone:
        """Get a specific zone.

        :param zone_id: The ID of the zone to get
        :param timeout: The timeout to use for this request
        """
        req = f"/api/zones/{zone_id}"
        data, _ = self._get_json(req, timeout)
        return Zone.from_dict(data)

    def set_zone(self, zone: Zone,
                 timeout=DEFAULT_TIMEOUT):
        """Update or create a zone. If the Zone doesn't exist, the zone.id
        must be None. An initialized Zone with an ID will be returned.

        :param zone: A Zone object
        :param timeout: The timeout to use for this request
        :return: Zone, initialized with an ID
        """
        req = "/api/zones"
        data = self._post_codec(req, timeout, zone)
        new_zone = Zone.from_dict(data)
        return new_zone

    def delete_zone(self, zone_id: int,
                    timeout=DEFAULT_TIMEOUT):
        """Deletes a zone with the given ID.

        :param zone_id: The ID of the zone to delete
        :param timeout: The timeout to use for this request
        """
        req = f"/api/zones/{zone_id}"
        self._delete(req, timeout)

    def is_zone_read_only(self, zone_id: int, timeout: float = DEFAULT_TIMEOUT) -> bool:
        """Checks if a zone is read-only. This refers to the zone itself, not its
        alarms or any other attached resources.

        :param zone_id: The ID of the zone to check
        :param timeout: The timeout to use for this request
        :return: True if the zone is read-only
        """
        req = f"/api/zones/{zone_id}"
        resp = self._options(req, timeout)

        # Parse the Allow header, which is a string of comma-separated method names
        allow_header = resp.headers["Allow"]
        allowed_methods = allow_header.split(",")
        allowed_methods = [m.strip().upper() for m in allowed_methods]

        return "POST" not in allowed_methods



from typing import Optional
from enum import Enum

from dataclasses import dataclass, field

from .base_codecs import Codec


@dataclass
class StreamConfiguration(Codec):
    """Describes a video stream that BrainFrame may connect to and analyze.
    """

    class ConnType(Enum):
        IP_CAMERA = "ip_camera"
        """A network camera that uses RTSP or HTTP"""
        WEBCAM = "webcam"
        """A webcam (usually USB)"""
        FILE = "file"
        """An uploaded video file"""

        @classmethod
        def values(cls):
            return [v.value for v in cls]

    name: str
    """The human-readable name of the video stream"""

    premises_id: Optional[int]
    """The ID of the premises that this stream is a part of, or None if
    this stream is not part of a premises
    """

    connection_type: ConnType
    """The type of stream this configuration is describing"""

    connection_options: dict
    """Contains configuration describing how this stream can be connected
    to. The contents of this dict will depend on the connection type.

    ConnType.IP_CAMERA:
        ``url``: The URL of the IP camera to connect to

        ``pipeline`` (optional): A custom GStreamer pipeline to connect to
        the IP camera with

    ConnType.FILE:
        ``storage_id``: The storage ID of the video file to stream

        ``transcode`` (optional): If True, the video file will be
        transcoded before being streamed. By default, this value is True.

        ``pipeline`` (optional): A custom GStreamer pipeline to connect to
        the hosted stream of this video file with

    ConnType.WEBCAM:
        ``device_id``: The Video4Linux device ID of the webcam
    """

    runtime_options: dict
    """Key-value pairs of configuration information that changes the
    runtime behavior of the video stream from its defaults
    """

    metadata: dict = field(default_factory=dict)
    """Key-value pairs containing some custom user-defined set of data to
    further describe the stream
    """

    id: Optional[int] = None
    """The unique ID of this stream"""

    def to_dict(self):
        d = dict(self.__dict__)
        d["connection_type"] = self.connection_type.value
        return d

    @staticmethod
    def from_dict(d):
        connection_t = StreamConfiguration.ConnType(d["connection_type"])
        return StreamConfiguration(name=d["name"],
                                   id=d["id"],
                                   connection_type=connection_t,
                                   connection_options=d["connection_options"],
                                   runtime_options=d["runtime_options"],
                                   metadata=d["metadata"],
                                   premises_id=d["premises_id"])



from dataclasses import dataclass
from typing import List

from .base_codecs import Codec


@dataclass
class OAuth2Info(Codec):
    """Information necessary to start an OAuth2 flow on behalf of the BrainFrame server
    """
    domain: str
    """The domain of the OAuth2 server"""

    client_id: str
    """The OAuth2 client ID"""

    scopes: List[str]
    """The OAuth2 scopes that tokens must have access to"""

    def to_dict(self) -> dict:
        return dict(self.__dict__)

    @staticmethod
    def from_dict(d: dict) -> "OAuth2Info":
        return OAuth2Info(
            domain=d["domain"],
            client_id=d["client_id"],
            scopes=d["scopes"],
        )



from typing import List, Optional

from brainframe.api.bf_codecs import Encoding
from .base_stub import BaseStub, DEFAULT_TIMEOUT


class EncodingStubMixIn(BaseStub):
    """Provides stubs to call APIs that handle encodings."""

    def get_encodings(self, identity_id: Optional[int] = None,
                      class_name: Optional[str] = None,
                      timeout=DEFAULT_TIMEOUT) -> List[Encoding]:
        """Gets all encodings from the server that match the given filters.

        :param identity_id: If specified, only encodings attached to this
            identity will be returned
        :param class_name: If specified, only encodings for the given class
            name will be returned
        :param timeout: The timeout to use for this request
        :return: All encodings that match this filter
        """
        req = f"/api/encodings"
        params = {}
        if identity_id is not None:
            params["identity_id"] = identity_id
        if class_name is not None:
            params["class_name"] = class_name

        encodings, _ = self._get_json(req, timeout, params=params)
        encodings = [Encoding.from_dict(e) for e in encodings]

        return encodings

    def get_encoding_class_names(self, identity_id: Optional[int] = None,
                                 timeout=DEFAULT_TIMEOUT) \
            -> List[str]:
        """Get all unique class names for encodings that match the given
        filter.

        :param identity_id: If specified, only class names for encodings
            attached to this identity will be returned
        :param timeout: The timeout to use for this request
        :return: All class names from encodings that match this filter
        """
        req = f"/api/encodings"
        params = {"fields": "class_name"}
        if identity_id is not None:
            params["identity_id"] = identity_id

        encodings, _ = self._get_json(req, timeout, params=params)
        class_names = [e["class_name"] for e in encodings]

        return class_names

    def get_encoding(self, encoding_id,
                     timeout=DEFAULT_TIMEOUT) -> Encoding:
        """Get the encoding with the given ID.

        :param encoding_id: The encoding ID to get an encoding for
        :param timeout: The timeout to use for this request
        :return: The corresponding encoding
        """
        req = f"/api/encodings/{encoding_id}"

        encoding, _ = self._get_json(req, timeout)
        return Encoding.from_dict(encoding)

    def delete_encoding(self, encoding_id,
                        timeout=DEFAULT_TIMEOUT):
        """Deletes the encoding with the given ID.

        :param encoding_id: The ID of the encoding to delete
        :param timeout: The timeout to use for this request
        """
        req = f"/api/encodings/{encoding_id}"
        self._delete(req, timeout)

    def delete_encodings(self, identity_id=None, class_name=None,
                         timeout=DEFAULT_TIMEOUT):
        """Deletes all encodings that match the given filter.

        :param identity_id: If specified, only encodings that are associated
            with this identity will be deleted
        :param class_name: If specified, only encodings that are for this class
            will be deleted
        :param timeout: The timeout to use for this request
        """
        req = f"/api/encodings"

        params = {}
        if identity_id is not None:
            params["identity_id"] = identity_id
        if class_name is not None:
            params["class_name"] = class_name

        self._delete(req, timeout, params)



from typing import Optional

from brainframe.api.bf_codecs import CloudUserInfo
from brainframe.api.bf_errors import CloudUserNotFoundError

from .base_stub import BaseStub, DEFAULT_TIMEOUT


class CloudUsersStubMixIn(BaseStub):
    def get_current_cloud_user(self, timeout=DEFAULT_TIMEOUT) \
            -> Optional[CloudUserInfo]:
        """Gets information on the cloud user that's currently logged in.

        :param timeout: The timeout to use for this request
        :return: Information on the cloud user, or None if no user is logged in
        """
        req = f"/api/cloud_user"
        try:
            data, _ = self._get_json(req, timeout=timeout)
        except CloudUserNotFoundError:
            return None
        else:
            return CloudUserInfo.from_dict(data)



from enum import Enum
from typing import Optional

from dataclasses import dataclass

from .base_codecs import Codec
from .detection_codecs import Attribute


condition_test_map = {'>': "Greater than",
                      '>=': "Greater than or equal to",
                      '<': "Less than",
                      '<=': "Less than or equal to",
                      "=": "Exactly",
                      "!=": "Not Equal To"}


class IntersectionPointType(Enum):
    """The point on a detection that must be inside a zone for the detection to
    count as being inside the zone. The most commonly used intersection point
    is BOTTOM, which counts a detection as being inside a zone if the bottom
    center point of the detection in the zone.
    """

    BOTTOM = "bottom"
    TOP = "top"
    LEFT = "left"
    RIGHT = "right"
    CENTER = "center"

    @classmethod
    def values(cls):
        return [v.value for v in cls]


@dataclass
class ZoneAlarmCountCondition(Codec):
    """A condition that must be met for an alarm to go off. Compares the number
    of objects in a zone to some number.
    """

    class TestType(Enum):
        """Defines the way a count condition compares the actual value to the
        alarm's test value.
        """

        GREATER_THAN = ">"
        LESS_THAN = "<"
        EQUAL_TO = "="
        NOT_EQUAL_TO = "!="

        @classmethod
        def values(cls):
            return [v.value for v in cls]

    test: TestType
    """The way that the check value will be compared to the actual count
    """

    check_value: int
    """The value to test the actual count against"""

    with_class_name: str
    """The class name of the objects to count"""

    with_attribute: Optional[Attribute]
    """If provided, only objects that have this attribute value will be
    counted.
    """

    window_duration: float
    """The sliding window duration for this condition"""

    window_threshold: float
    """The portion of time during the sliding window duration that this
    condition must be true in order for the associated alarm to trigger
    """

    intersection_point: IntersectionPointType
    """The point in each detection that must be within the zone in order
    for that detection to be counted as in that zone
    """

    id: Optional[int] = None
    """A unique identifier"""

    def __repr__(self):
        condition_str = condition_test_map[self.test.value]
        attr = self.with_attribute
        attr = attr.value + ' ' if attr else ''
        return f"{condition_str} {self.check_value} " \
               f"{attr}{self.with_class_name}(s) "

    def to_dict(self) -> dict:
        d = dict(self.__dict__)
        d["test"] = self.test.value
        d["intersection_point"] = self.intersection_point.value
        if self.with_attribute is not None:
            d["with_attribute"] = self.with_attribute.to_dict()

        return d

    @staticmethod
    def from_dict(d: dict):
        intersection_point = IntersectionPointType(
            d["intersection_point"])
        # with_attribute is an optional parameter
        with_attribute = None
        if d["with_attribute"] is not None:
            with_attribute = Attribute.from_dict(d["with_attribute"])
        test = ZoneAlarmCountCondition.TestType(d["test"])

        return ZoneAlarmCountCondition(
            test=test,
            check_value=d["check_value"],
            with_class_name=d["with_class_name"],
            with_attribute=with_attribute,
            window_duration=d["window_duration"],
            window_threshold=d["window_threshold"],
            intersection_point=intersection_point,
            id=d["id"])


@dataclass
class ZoneAlarmRateCondition(Codec):
    """A condition that must be met for an alarm to go off. Compares the rate
    of change in the count of some object against a test value.
    """

    class TestType(Enum):
        """Defines the way a rate condition compares the actual rate value to
        the alarm's test value.
        """

        GREATER_THAN_OR_EQUAL_TO = ">="
        LESS_THAN_OR_EQUAL_TO = "<="

        @classmethod
        def values(cls):
            return [v.value for v in cls]

    class DirectionType(Enum):
        """Defines the direction of flow that a rate condition pertains to."""

        ENTERING = "entering"
        EXITING = "exiting"
        ENTERING_OR_EXITING = "entering_or_exiting"

        @classmethod
        def values(cls):
            return [v.value for v in cls]

    _direction_map = {DirectionType.ENTERING: "entered",
                      DirectionType.EXITING: "exited",
                      DirectionType.ENTERING_OR_EXITING: "entered or exited"}

    test: TestType
    """The way that the change value will be compared to the actual rate"""

    duration: float
    """The time in seconds for this rate change to occur"""

    change: float
    """The rate change value to compare the actual rate value against"""

    direction: DirectionType
    """The direction of flow this condition tests for"""

    with_class_name: str
    """The class name of the objects to measure rate of change for"""

    with_attribute: Optional[Attribute]
    """If provided, only objects that have this attribute will be counted
    in the rate calculation
    """

    intersection_point: IntersectionPointType
    """The point in each detection that must be within the zone in order
    for that detection to be counted as in that zone
    """

    id: int = None
    """A unique identifier"""

    def __repr__(self):
        condition_str = condition_test_map[self.test.value]
        direction_str = self._direction_map[self.direction]
        return f"{condition_str} {self.change} {self.with_class_name}(s) " \
               f"{direction_str} within {self.duration} seconds"

    def to_dict(self) -> dict:
        d = dict(self.__dict__)
        d["intersection_point"] = self.intersection_point.value
        if self.with_attribute is not None:
            d["with_attribute"] = self.with_attribute.to_dict()

        d["direction"] = self.direction.value
        d["test"] = self.test.value

        return d

    @staticmethod
    def from_dict(d: dict):
        intersection_point = IntersectionPointType(
            d["intersection_point"])
        # with_attribute is an optional parameter
        with_attribute = None
        if d["with_attribute"] is not None:
            with_attribute = Attribute.from_dict(d["with_attribute"])
        test = ZoneAlarmRateCondition.TestType(d["test"])

        return ZoneAlarmRateCondition(
            test=test,
            duration=d["duration"],
            change=d["change"],
            direction=ZoneAlarmRateCondition.DirectionType(d["direction"]),
            with_class_name=d["with_class_name"],
            with_attribute=with_attribute,
            intersection_point=intersection_point,
            id=d["id"])



from typing import List, Optional

from dataclasses import dataclass

from .base_codecs import Codec
from .condition_codecs import ZoneAlarmCountCondition, ZoneAlarmRateCondition


@dataclass
class ZoneAlarm(Codec):
    """This is the configuration for an alarm."""

    name: str
    """A friendly name for the zone alarm"""

    count_conditions: List[ZoneAlarmCountCondition]
    """All count conditions for this alarm"""

    rate_conditions: List[ZoneAlarmRateCondition]
    """All rate conditions for this alarm"""

    use_active_time: bool
    """If True, the alarm will only be triggered when the current time is
    between the active_start_time and active_end_time.
    """

    active_start_time: str
    """The time of day where this alarm starts being active, in the format
    "hh:mm:ss"
    """

    active_end_time: str
    """The time of day where this alarm starts being active, in the format
    "hh:mm:ss"
    """

    id: Optional[int] = None
    """A unique identifier"""

    zone_id: Optional[int] = None
    """The ID of the zone this alarm is associated with"""

    stream_id: Optional[int] = None
    """The ID of the stream the associated zone is in"""

    def to_dict(self):
        d = dict(self.__dict__)
        d["count_conditions"] = [ZoneAlarmCountCondition.to_dict(cond)
                                 for cond in self.count_conditions]
        d["rate_conditions"] = [ZoneAlarmRateCondition.to_dict(cond)
                                for cond in self.rate_conditions]

        return d

    @staticmethod
    def from_dict(d):
        count_conditions = [ZoneAlarmCountCondition.from_dict(cond)
                            for cond in d["count_conditions"]]
        rate_conditions = [ZoneAlarmRateCondition.from_dict(cond)
                           for cond in d["rate_conditions"]]

        return ZoneAlarm(name=d["name"],
                         id=d["id"],
                         count_conditions=count_conditions,
                         rate_conditions=rate_conditions,
                         use_active_time=d["use_active_time"],
                         active_start_time=d["active_start_time"],
                         active_end_time=d["active_end_time"],
                         zone_id=d["zone_id"],
                         stream_id=d["stream_id"])


@dataclass
class Alert(Codec):
    """This is sent when an Alarm has been triggered."""

    alarm_id: int
    """The ID of the alarm that this alert came from"""

    zone_id: int
    """The ID of the zone this alert pertains to"""

    stream_id: int
    """The ID of the stream this alert pertains to"""

    start_time: float
    """When the event started happening, in Unix Time (seconds)"""

    end_time: Optional[float]
    """When the event stopped happening, in Unix Time (seconds), or None
    if the alert is ongoing.
    """

    verified_as: Optional[bool]
    """
    - If True, then this alert was labeled by a person as legitimate
    - If False, then this alert was labeled by a person as a false alarm
    - If None, then this alert has not been reviewed by a person
    """

    id: Optional[int] = None
    """A unique identifier"""

    def to_dict(self):
        d = dict(self.__dict__)
        return d

    @staticmethod
    def from_dict(d):
        return Alert(id=d["id"],
                     alarm_id=d["alarm_id"],
                     zone_id=d["zone_id"],
                     stream_id=d["stream_id"],
                     start_time=d["start_time"],
                     end_time=d["end_time"],
                     verified_as=d["verified_as"])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['api', 'api.bf_codecs', 'api.stubs']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.16,<2.0', 'pillow>=7.2,<9.0', 'requests>=2.24.0,<3.0.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.7,<0.8']}

setup_kwargs = {
    'name': 'brainframe-api',
    'version': '0.29.1',
    'description': 'Provides a Python wrapper around the BrainFrame REST API.',
    'long_description': '=====================\nBrainFrame Python API\n=====================\n\n.. image:: https://readthedocs.org/projects/brainframe-python-api/badge/?version=latest\n   :target: https://brainframe-python-api.readthedocs.io/en/latest/?badge=latest\n   :alt: Documentation Status\n\n.. image:: https://circleci.com/gh/aotuai/brainframe-python.svg?style=svg\n   :target: https://circleci.com/gh/aotuai/brainframe-python\n   :alt: Publish Packages CI Status\n\nThis library is a Python wrapper around the BrainFrame REST API. It allows for\neasy interaction with a BrainFrame server.\n\n.. code-block:: python\n\n   from brainframe.api import BrainFrameAPI, bf_codecs\n\n   # Connect to the server\n   api = BrainFrameAPI("http://localhost")\n   # Create a new IP camera stream\n   stream_config = api.set_stream_configuration(\n       bf_codecs.StreamConfiguration(\n           name="New Stream",\n           connection_type=bf_codecs.ConnType.IP_CAMERA,\n           connection_options={"url": "rtsp://192.168.1.100"},\n           runtime_options={},\n       ))\n   api.start_analyzing(stream_config.id)\n   # Get results\n   analysis_results = api.get_latest_zone_statuses()\n\n\nInstallation\n------------\n\nThe BrainFrame Python API is available on PyPI and can be installed with pip.\nInstall the version of the library that matches the version of BrainFrame that\nyou are using. For example, if you are using BrainFrame version 0.26.1:\n\n.. code-block:: bash\n\n   pip3 install brainframe-api\n\nDocumentation\n-------------\n\nDocumentation for this library is available on `ReadTheDocs`_.\n\n.. _`ReadTheDocs`:\n   https://brainframe-python-api.readthedocs.io/en/latest/\n\n',
    'author': 'Aotu',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/aotuai/brainframe_python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from enum import Enum
from typing import Dict, List, Any

from dataclasses import dataclass

from .base_codecs import Codec


@dataclass
class CapsuleOption(Codec):
    """A single configuration option for a capsule. Defines what type of option
    it is and its potential values.

    There are two kinds of capsule options. Stream capsule options apply only
    to the stream they are attached to. Global capsule options apply to all
    streams, but are overridden by stream capsule options.
    """

    class Type(Enum):
        """The data type of a capsule option"""

        FLOAT = "float"
        INT = "int"
        ENUM = "enum"
        BOOL = "bool"

        @classmethod
        def values(cls):
            return [v.value for v in cls]

    type: Type
    """The data type of this option's value"""

    default: Any
    """The default value for this option"""

    constraints: dict
    """Describes the range of valid values for this option. The content of
    this dict depends on the type field.
    
    OptionType.FLOAT:
        ``max_val``: The maximum valid float value
        
        ``min_val``: The minimum valid float value
    
    OptionType.INT:
        ``max_val``: The maximum valid int value
        
        ``min_val``: The minimum valid int value
    
    OptionType.ENUM:
        ``choices``: A list of strings. The option's value must be one of
        these strings.
    
    OptionType.BOOL:
        This object has no constraints.
    """

    description: str
    """A human-readable description of the capsule's capabilities"""

    def to_dict(self):
        d = dict(self.__dict__)
        d["type"] = self.type.value
        return d

    @staticmethod
    def from_dict(d):
        type_ = CapsuleOption.Type(d["type"])
        return CapsuleOption(type=type_,
                             default=d["default"],
                             constraints=d["constraints"],
                             description=d["description"])


@dataclass
class NodeDescription(Codec):
    """A description of a DetectionNode, used by capsules to define what kinds
    of inputs and outputs a capsule uses.
    """

    class Size(Enum):
        """Describes the amount of DetectionNodes a capsule takes as input or
        provides as output.
        """

        NONE = "none"
        """Input: The capsule takes nothing as input, like for an object
        detector.

        Output: Capsules cannot have a NONE output.
        """
        SINGLE = "single"
        """Input: The capsule takes a single DetectionNode as input, like for a
        classifier.

        Output: The capsule provides a single modified DetectionNode as output,
        like for a classifier.
        """
        ALL = "all"
        """Input: The capsule takes all instances of a class as input, like for
        a tracker.

        Output: The capsule provides all instances of a class as output, like
        for a detector.
        """

        @classmethod
        def values(cls):
            return [v.value for v in cls]

    size: Size
    """Describes the amount of DetectionNodes the node either takes in as
    input or provides as output
    """

    detections: List[str]
    """A list of detection class names, like “person” or “vehicle”. A
    DetectionNode that meets this description must have a class name that
    is present in this list.
    """

    attributes: Dict[str, List[str]]
    """Key-value pairs whose key is the classification type and whose value
    is a list of possible attributes. A DetectionNode that meets this
    description must have a classification for each classification type
    listed here.
    """

    encoded: bool
    """If True, the DetectionNode must be encoded to meet this description
    """

    tracked: bool
    """If True, the DetectionNode must be tracked to meet this description
    """

    extra_data: List[str]
    """A list of keys in a NodeDescription's extra_data. A DetectionNode
    that meets this description must have extra data for each name listed
    here.
    """

    def to_dict(self):
        d = dict(self.__dict__)
        d["size"] = self.size.value
        return d

    @staticmethod
    def from_dict(d):
        size = NodeDescription.Size(d["size"])
        return NodeDescription(
            size=size,
            detections=d["detections"],
            attributes=d["attributes"],
            encoded=d["encoded"],
            tracked=d["tracked"],
            extra_data=d["extra_data"])


@dataclass
class Capsule(Codec):
    """Metadata on a loaded capsule."""

    name: str
    """The name of the capsule"""

    version: int
    """The capsule's version"""

    description: str
    """A human-readable description of what the capsule does"""

    input_type: NodeDescription
    """Describes the type of inference data that this capsule takes as input
    """

    output_type: NodeDescription
    """Describes the type of inference data that this capsule produces"""

    capability: NodeDescription
    """A NodeDescription which describes what this capsule does to its
    input. It is the difference between the input and output
    NodeDescriptions. This field is useful for inspecting a capsule to find
    what it can do.
    """

    options: Dict[str, CapsuleOption]
    """A dict describing the configurable options of this capsule"""

    def to_dict(self):
        return {
            "name": self.name,
            "version": self.version,
            "description": self.description,
            "input_type": self.input_type.to_dict(),
            "output_type": self.output_type.to_dict(),
            "capability": self.capability.to_dict(),
            "options": {key: val.to_dict()
                        for key, val in self.options.items()},
        }

    @staticmethod
    def from_dict(d):
        return Capsule(
            name=d["name"],
            version=d["version"],
            description=d["description"],
            input_type=NodeDescription.from_dict(d["input_type"]),
            output_type=NodeDescription.from_dict(d["output_type"]),
            capability=NodeDescription.from_dict(d["capability"]),
            options={key: CapsuleOption.from_dict(val)
                     for key, val in d["options"].items()},
        )



import re
from typing import Optional

kind_to_error_type = {}
"""Maps error kinds to their corresponding error type."""


def _server_origin_error(kind: Optional[str] = None):
    """A class decorator. Registers the given class so that it will be used
    when an error with a kind equal to the class's name is provided as a
    response. Used only for Errors that can originate on the BrainFrame server.

    :param kind: If provided, this class will be associated with this kind
        instead of using the class's name as the kind
    """

    def wrapper(cls):
        global kind_to_error_type
        nonlocal kind

        if kind is None:
            kind = cls.__name__
        kind_to_error_type[kind] = cls
        cls.kind = kind
        return cls

    return wrapper


class BaseAPIError(Exception):
    """All API errors subclass this error."""
    # This is set by the decorator
    kind = None

    def __init__(self, description):
        self.description = description
        super().__init__(f"{self.kind}: {description}")

    @property
    def pretty_name(self):
        name = self.__class__.__name__

        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1 \2', name)
        s2 = re.sub('([a-z0-9])([A-Z])', r'\1 \2', s1).capitalize()
        s3 = s2.rsplit(" error")[0]

        return s3


@_server_origin_error()
class UnknownError(BaseAPIError):
    """Something unexpected happened. The server may be in an invalid state."""

    def __init__(self, description, status_code=None):
        super().__init__(description)
        self.status_code = status_code


@_server_origin_error()
class StreamConfigNotFoundError(BaseAPIError):
    """A StreamConfiguration specified by the client could not be found."""


@_server_origin_error()
class ZoneNotFoundError(BaseAPIError):
    """A Zone specified by the client could not be found."""


@_server_origin_error()
class PremisesNotFoundError(BaseAPIError):
    """A Zone specified by the client could not be found."""


@_server_origin_error()
class ZoneNotDeletableError(BaseAPIError):
    """A client tried to delete a default Zone"""


@_server_origin_error()
class ZoneNotEditableError(BaseAPIError):
    """A client tried to edit a read-only Zone"""


@_server_origin_error()
class AlertNotFoundError(BaseAPIError):
    """An Alert specified by the client could not be found."""


@_server_origin_error()
class InvalidSyntaxError(BaseAPIError):
    """The syntax of the request could not be parsed."""


@_server_origin_error()
class InvalidFormatError(BaseAPIError):
    """The request was parsed, but some value within the request is invalid."""


@_server_origin_error("NotImplementedError")
class NotImplementedInAPIError(BaseAPIError):
    """The client requested something that is not currently implemented."""


@_server_origin_error()
class StreamNotOpenedError(BaseAPIError):
    """A stream failed to open when it was required to."""


@_server_origin_error()
class DuplicateStreamSourceError(BaseAPIError):
    """There was an attempted to create a stream configuration with the same
    source as an existing one.
    """


@_server_origin_error()
class DuplicateZoneNameError(BaseAPIError):
    """There was an attempt to make a zone with the same name as another zone
    within the same stream.
    """


@_server_origin_error()
class DuplicateIdentityNameError(BaseAPIError):
    """There was an attempt to create a new identity with the same name as
    another identity.
    """


@_server_origin_error()
class NoDetectorForClassError(BaseAPIError):
    """There was an attempt to use a class name that is not detectable."""


@_server_origin_error()
class NoEncoderForClassError(BaseAPIError):
    """There was an attempt to create an identity for a class that is not
    encodable.
    """


@_server_origin_error()
class IdentityNotFoundError(BaseAPIError):
    """An identity specified by the client could not be found."""


@_server_origin_error()
class ImageNotFoundForIdentityError(BaseAPIError):
    """An image specified by the client could not be found for the specified
    identity.
    """


@_server_origin_error()
class InvalidImageTypeError(BaseAPIError):
    """An image could not be decoded by OpenCV"""


@_server_origin_error()
class AnalysisLimitExceededError(BaseAPIError):
    """There was an attempt to start analysis on a stream, but the maximum
    amount of streams that may have analysis run on them at once has already
    been reached.
    """


@_server_origin_error()
class NoDetectionsInImageError(BaseAPIError):
    """There was an attempt to encode an image with no objects of the given
    class in the frame.
    """


@_server_origin_error()
class TooManyDetectionsInImageError(BaseAPIError):
    """There was an attempt to encode an image with more than one object of the
    given class in the frame, causing ambiguity on which one to encode.
    """


@_server_origin_error()
class ImageAlreadyEncodedError(BaseAPIError):
    """There was an attempt to encode an image that has already been encoded
    for a given identity and a given class.
    """


@_server_origin_error()
class DuplicateVectorError(BaseAPIError):
    """There was an attempt to add a vector that already exists for the given
    identity and class.
    """


@_server_origin_error()
class FrameNotFoundForAlertError(BaseAPIError):
    """There was an attempt to get a frame for an alert that has no frame."""


@_server_origin_error("PluginNotFoundError")
class CapsuleNotFoundError(BaseAPIError):
    """There was an attempt to reference a capsule that does not exist."""


@_server_origin_error("InvalidPluginOptionError")
class InvalidCapsuleOptionError(BaseAPIError):
    """The provided capsule options do not work for the given capsule. This
    could be because the option does not exist or the value for that option
    doesn't fit the constraints.
    """


@_server_origin_error()
class StorageNotFoundError(BaseAPIError):
    """There was an attempt to access a storage object that does not exist."""


@_server_origin_error()
class ZoneAlarmNotFoundError(BaseAPIError):
    """There was an attempt to access a zone alarm that does not exist."""


@_server_origin_error()
class InvalidRuntimeOptionError(BaseAPIError):
    """There was an attempt to set a runtime option that is not supported or is
    of the wrong type.
    """


@_server_origin_error()
class EncodingNotFoundError(BaseAPIError):
    """There was an attempt to access an encoding that does not exist."""


@_server_origin_error()
class UnauthorizedError(BaseAPIError):
    """There was an attempt to access the API without proper authorization."""


@_server_origin_error()
class InvalidSessionError(BaseAPIError):
    """There was an attempt to access the API with an invalid session ID,
    either because the session expired or no session with that ID has ever
    existed. The client should authenticate again to get a new session.
    """


@_server_origin_error()
class VectorTooLongError(BaseAPIError):
    """The provided encoding vector is longer than the maximum allowed
    length.
    """


@_server_origin_error()
class UserNotFoundError(BaseAPIError):
    """There was an attempt to access a user by ID that does not exist."""


@_server_origin_error()
class InsufficientRoleError(BaseAPIError):
    """A user attempted an operation that they don't have permission to do."""


@_server_origin_error()
class DuplicateUsernameError(BaseAPIError):
    """The requested username already exists."""


@_server_origin_error()
class AdminMustExistError(BaseAPIError):
    """There was an attempt to delete the only remaining admin account."""


@_server_origin_error()
class LicenseRequiredError(BaseAPIError):
    """There was an attempt to access a resource that requires an active
    license while no valid license is loaded.
    """


@_server_origin_error()
class LicenseExpiredError(BaseAPIError):
    """There was an attempt to upload a license that is expired."""


@_server_origin_error()
class LicenseInvalidError(BaseAPIError):
    """There was an attempt to upload a license that is in an invalid format.
    """


@_server_origin_error()
class RemoteConnectionError(BaseAPIError):
    """The server encountered an error while connecting to a remote resource
    that is required for the requested operation.
    """


@_server_origin_error()
class InvalidCapsuleError(BaseAPIError):
    """The provided capsule could not be loaded."""


@_server_origin_error()
class IncompatibleCapsuleError(BaseAPIError):
    """The provided capsule is not compatible with this version of BrainFrame.
    """


@_server_origin_error()
class UnauthorizedTokensError(BaseAPIError):
    """The provided tokens do not correctly authorize a BrainFrame Cloud user"""


@_server_origin_error()
class CloudUserNotFoundError(BaseAPIError):
    """No cloud user is logged in"""


class ServerNotReadyError(BaseAPIError):
    """The client was able to communicate with the server, but the server had
    not completed startup or was in an invalid state"""



from typing import Tuple

from brainframe.api.bf_codecs import CloudTokens, CloudUserInfo, LicenseInfo
from .base_stub import BaseStub, DEFAULT_TIMEOUT


class CloudTokensStubMixin(BaseStub):
    def set_cloud_tokens(self, cloud_tokens: CloudTokens,
                         timeout=DEFAULT_TIMEOUT) -> Tuple[CloudUserInfo, LicenseInfo]:
        """Authorizes the server against BrainFrame Cloud using the provided tokens.

        :param cloud_tokens: The tokens to use for authorization
        :param timeout: the timeout to use for this request
        :return: Info on the BrainFrame Cloud user that corresponds to these tokens
        """
        req = "/api/cloud_tokens"
        login_result = self._put_codec(req, timeout, cloud_tokens)

        cloud_user_info = CloudUserInfo.from_dict(login_result["cloud_user_info"])
        license_info = LicenseInfo.from_dict(login_result["license_info"])

        return cloud_user_info, license_info



import json
from pathlib import Path
from time import time
from typing import Dict, List, Optional, BinaryIO, Union

import requests.exceptions

from brainframe.api.bf_codecs import Capsule
from .base_stub import DEFAULT_TIMEOUT
from .storage import StorageStubMixin


class CapsuleStubMixin(StorageStubMixin):
    """Provides stubs to call APIs to inspect and configure capsules."""

    def get_capsule(self, name,
                    timeout=DEFAULT_TIMEOUT) -> Capsule:
        """
        :param name: The name of the capsule to get
        :param timeout: The timeout to use for this request
        :return: Capsule with the given name
        """
        req = f"/api/plugins/{name}"
        capsule, _ = self._get_json(req, timeout)
        return Capsule.from_dict(capsule)

    def get_capsules(self, timeout=DEFAULT_TIMEOUT) -> List[Capsule]:
        """
        :param timeout: The timeout to use for this request
        :return: All available capsules
        """
        req = "/api/plugins"
        capsules, _ = self._get_json(req, timeout)
        return [Capsule.from_dict(d) for d in capsules]

    def load_capsule(self, storage_id: int,
                     source_path: Optional[Path] = None,
                     timeout: float = DEFAULT_TIMEOUT) -> Capsule:
        """Loads and initializes a capsule from capsule data in storage.

        :param storage_id: The ID of the raw capsule data in storage
        :param source_path: If available, the path to the source code on the
            development machine can be provided. Doing so allows stack traces
            to point to the correct source location.
        :param timeout: The timeout to use for this request
        :return: The loaded capsule
        """
        req = f"/api/plugins"
        req_object = {
            "storage_id": storage_id,
            "dev_options": {
                "source_path": str(source_path),
            },
        }
        capsule = self._put_json(req, timeout, json.dumps(req_object))
        return Capsule.from_dict(capsule)

    def upload_and_load_capsule(self, data: Union[bytes, BinaryIO],
                                source_path: Optional[Path] = None,
                                timeout: float = DEFAULT_TIMEOUT) -> Capsule:
        """Uploads capsule data to storage, then loads and initializes a
        capsule from it. This is a utility method that combines the
        functionality of `new_storage` and `load_capsule`.

        :param data: The data to store, either as bytes or as a file-like
        :param source_path: If available, the path to the source code on the
            development machine can be provided. Doing so allows stack traces
            to point to the correct source location.
        :param timeout: The timeout to use for each request
        :return: The loaded capsule
        """
        storage_id = self.new_storage(data,
                                      mime_type="application/octet-stream",
                                      timeout=timeout)

        return self.load_capsule(storage_id,
                                 source_path,
                                 timeout=timeout)

    def unload_capsule(self, capsule_name: str,
                       timeout: float = DEFAULT_TIMEOUT) -> None:
        """Unloads a capsule and deletes its data from storage. This method may
        only be used with capsules that were loaded through the REST API.

        :param capsule_name: The name of the capsule to unload
        :param timeout: The timeout to use for this request
        """
        req = f"/api/plugins/{capsule_name}"
        self._delete(req, timeout)

    def get_capsule_option_vals(self, capsule_name, stream_id=None,
                                timeout=DEFAULT_TIMEOUT) \
            -> Dict[str, object]:
        """Gets the current values for every capsule option. See the
        documentation for the CapsuleOption codec for more info about global
        and stream level options and how they interact.

        :param capsule_name: The capsule to find options for
        :param stream_id: The ID of the stream. If this value is None, then the
            global options are returned for that capsule
        :param timeout: The timeout to use for this request
        :return: A dict where the key is the option name and the value is the
            option's current value
        """
        if stream_id is None:
            req = f"/api/plugins/{capsule_name}/options"
        else:
            req = f"/api/streams/{stream_id}/plugins/{capsule_name}/options"
        capsule_option_vals, _ = self._get_json(req, timeout)

        return capsule_option_vals

    def set_capsule_option_vals(self, *, capsule_name, stream_id=None,
                                option_vals: Dict[str, object],
                                timeout=DEFAULT_TIMEOUT):
        """Sets option values for a capsule.

        :param capsule_name: The name of the capsule whose options to set
        :param stream_id: The ID of the stream, if these are stream-level
            options. If this value is None, then the global options are set
        :param option_vals: A dict where the key is the name of the option to
            set, and the value is the value to set that option to
        :param timeout: The timeout to use for this request
        """
        if stream_id is None:
            req = f"/api/plugins/{capsule_name}/options"
        else:
            req = f"/api/streams/{stream_id}/plugins/{capsule_name}/options"

        option_values_json = json.dumps(option_vals)
        self._put_json(req, timeout, option_values_json)

    def patch_capsule_option_vals(self, *, capsule_name, stream_id=None,
                                  option_vals: Dict[str, object],
                                  timeout=DEFAULT_TIMEOUT):
        """Patches option values for a capsule. Only the provided options are
        changed. To unset an option, provide that option with a value of None.

        :param capsule_name: The name of the capsule whose options to set
        :param stream_id: The ID of the stream, if these are stream-level
            options. If this value is None, then the global options are set
        :param option_vals: A dict where the key is the name of the option to
            set, and the value is the value to set that option to
        :param timeout: The timeout to use for this request
        """
        if stream_id is None:
            req = f"/api/plugins/{capsule_name}/options"
        else:
            req = f"/api/streams/{stream_id}/plugins/{capsule_name}/options"

        option_values_json = json.dumps(option_vals)
        self._patch_json(req, timeout, option_values_json)

    def is_capsule_active(self, capsule_name, stream_id=None,
                          timeout=DEFAULT_TIMEOUT) -> bool:
        """Returns True if the capsule is active. If a capsule is not marked as
        active, it will not run. Like capsule options, this can be configured
        globally and on a per-stream level.

        :param capsule_name: The name of the capsule to get activity for
        :param stream_id: The ID of the stream, if you want the per-stream
            active setting
        :param timeout: The timeout to use for this request
        :return: True if the capsule is active
        """
        if stream_id is None:
            req = f"/api/plugins/{capsule_name}/active"
        else:
            req = f"/api/streams/{stream_id}/plugins/{capsule_name}/active"
        capsules_active, _ = self._get_json(req, timeout)
        return capsules_active

    def set_capsule_active(self, *, capsule_name, stream_id=None,
                           active: Optional[bool],
                           timeout=DEFAULT_TIMEOUT):
        """Sets whether or not the capsule is active. If a capsule is active, it
        will be run on frames.

        :param capsule_name: The name of the capsule to set activity for
        :param stream_id: The ID of the stream, if you want to set the
            per-stream active setting
        :param active: True if the capsule should be set to active
        :param timeout: The timeout to use for this request
        """
        if stream_id is None:
            req = f"/api/plugins/{capsule_name}/active"
        else:
            req = f"/api/streams/{stream_id}/plugins/{capsule_name}/active"

        active_json = json.dumps(active)

        self._put_json(req, timeout, active_json)



from .stub import BrainFrameAPI
from .stubs.base_stub import DEFAULT_TIMEOUT
from .status_receiver import StatusReceiver
from .stubs.zone_statuses import ZONE_STATUS_TYPE, ZONE_STATUS_STREAM_TYPE

__all__ = [
    "BrainFrameAPI",
    "DEFAULT_TIMEOUT",
    "StatusReceiver",
    "bf_errors",
    "bf_codecs",
    "ZONE_STATUS_TYPE",
    "ZONE_STATUS_STREAM_TYPE",
]
